/**************************************************************************
**
** This file is part of libqtcontacts-telepathy
**
** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
**
** Contact: Nokia Corporation (info@qt.nokia.com)
**
**
** GNU Lesser General Public License Usage
**
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this file.
** Please review the following information to ensure the GNU Lesser General
** Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** Other Usage
**
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at info@qt.nokia.com.
**
**************************************************************************/

#include "contactfetchrequest.h"

#include "engine.h"
#include "telepathy2qcontactconvertor.h"

///////////////////////////////////////////////////////////////////////////////////////////////////

QTelepathyContactFetchRequest::QTelepathyContactFetchRequest(QContactAbstractRequest *request,
                                                         QContactTelepathyEngine *engine,
                                                         QObject *parent)
    : QTelepathyBaseRequest<QContactFetchRequest>(request, engine, parent)
{
}

QTelepathyContactFetchRequest::~QTelepathyContactFetchRequest()
{
}

bool QTelepathyContactFetchRequest::start()
{
    const QString accountPath(engine()->accountPath());
    if (accountPath.isEmpty()) {
        reportError(QLatin1String("AccountPath was not supplied as parameter to QContactManager(\"telepathy\", parameters).\n"
                    "e.g.:\n"
                    "QMap<QString, QString>& parameters;\n"
                    "parameters[QLatin1String(\"AccountPath\")] = accountPath;\n"
                    "QContactManager *mTelepathyManager = new QContactManager(QLatin1String(\"telepathy\"), parameters);\n"
                    "Not able to search for contacts if account path is not supplied.\n"),
                    QContactManager::BadArgumentError);

    }
    Tp::Features wantedFeatures;

    wantedFeatures << Tp::Account::FeatureCore
                   << Tp::Account::FeatureProfile
                   << Tp::Account::FeatureProtocolInfo;

    m_account = Tp::Account::create(QLatin1String(TELEPATHY_ACCOUNT_MANAGER_BUS_NAME), accountPath);
    getChannel();
    return true;
}

void
QTelepathyContactFetchRequest::emitResult(QContactManager::Error error)
{
    engine()->updateContactFetchRequest(m_request, m_contacts, error,
                                       QContactAbstractRequest::FinishedState);
}

void
QTelepathyContactFetchRequest::getChannel()
{
    Tp::PendingChannel *pendingChannel = m_account->createAndHandleContactSearch();
    connect(pendingChannel, SIGNAL(finished(Tp::PendingOperation *)),
            this, SLOT(gotChannel(Tp::PendingOperation *)));
}

void
QTelepathyContactFetchRequest::gotChannel(Tp::PendingOperation*op)
{
    if(op->isError()) {
        reportError(QString::fromLatin1("Search connection channel not available:%1-%2 for account:%3. Not able to search.").
                arg(op->errorName()).
                arg(op->errorMessage()).
                arg(engine()->accountPath()), QContactManager::NotSupportedError);
    } else {
        Tp::PendingChannel *pendingChannel = qobject_cast<Tp::PendingChannel *>(op);
        m_channel = pendingChannel->channel();
        m_contactSearchChannel = Tp::ContactSearchChannelPtr::qObjectCast(m_channel);

        connect(m_contactSearchChannel->becomeReady(Tp::ContactSearchChannel::FeatureCore),SIGNAL(finished (Tp::PendingOperation*)),
                     SLOT(onChannelReady(Tp::PendingOperation*)));
    }
 }

void
QTelepathyContactFetchRequest::onChannelReady(Tp::PendingOperation *op)
{
    if (op->isError()) {
        reportError(QString::fromLatin1("Search connection channel not ready:%1-%2 for account:%3. Not able to search.").
                arg(op->errorName()).
                arg(op->errorMessage()).
                arg(engine()->accountPath()), QContactManager::UnspecifiedError);
    } else {
        // FIXME proper filterring
        // FIXME if no search string supplied, return roster list

        if (QContactFilter::ContactDetailFilter != m_request->filter().type()) {
            reportError(QString::fromLatin1("Only search - fetch using QContactFilter::ContactDetailFilter is supported."),
                    QContactManager::NotSupportedError);
            return;
        }

        const QContactDetailFilter detailFilter(m_request->filter());
        // FIXME WIP - filter
        m_searchOperation = m_contactSearchChannel->search(QLatin1String(""), detailFilter.value().toString());

        connect(m_contactSearchChannel.data(), SIGNAL(searchStateChanged(Tp::ChannelContactSearchState ,const QString &, const Tp::ContactSearchChannel::SearchStateChangeDetails &)),
                                              SLOT(onSearchStateChanged(Tp::ChannelContactSearchState ,const QString &, const Tp::ContactSearchChannel::SearchStateChangeDetails &)));

        connect(m_contactSearchChannel.data(), SIGNAL(searchResultReceived(const Tp::ContactSearchChannel::SearchResult &)),
                                              SLOT(onSearchResultReceived(const Tp::ContactSearchChannel::SearchResult&)));
    }
}

void
QTelepathyContactFetchRequest::onSearchStateChanged(Tp::ChannelContactSearchState state,const QString &error, const Tp::ContactSearchChannel::SearchStateChangeDetails &details)
{
    Q_UNUSED(details)
    if(state==Tp::ChannelContactSearchStateNotStarted) {
        qctWarn("Search not started");
    } else if (state==Tp::ChannelContactSearchStateInProgress) {
        engine()->updateRequestState(request(), QContactAbstractRequest::ActiveState);
    } else if (state==Tp::ChannelContactSearchStateCompleted) {
        emitResult(QContactManager::NoError);
    } else if (state==Tp::ChannelContactSearchStateFailed) {
        reportError(QString::fromLatin1("Search failed:%1 - for account:%3. Not able to search.").
                arg(error).
                arg(engine()->accountPath()), QContactManager::UnspecifiedError);
    }
}

void
QTelepathyContactFetchRequest::onSearchResultReceived(const Tp::ContactSearchChannel::SearchResult &result)
{
    QList<QContact> contacts;
    Telepathy2QContactConvertor::convertTelepathyContacts2QContacts(result, contacts, m_account->objectPath(), engine());
    engine()->updateContactFetchRequest(m_request, contacts, QContactManager::NoError,
                                       QContactAbstractRequest::QContactAbstractRequest::ActiveState);
    m_contacts << contacts;
}

///////////////////////////////////////////////////////////////////////////////////////////////////

#include "moc_contactfetchrequest.cpp"
