/**************************************************************************
**
** This file is part of libqtcontacts-telepathy
**
** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
**
** Contact: Nokia Corporation (info@qt.nokia.com)
**
**
** GNU Lesser General Public License Usage
**
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this file.
** Please review the following information to ensure the GNU Lesser General
** Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** Other Usage
**
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at info@qt.nokia.com.
**
**************************************************************************/

#include <qcontactdetails.h>

#include "engine.h"
#include "engine_p.h"

#include "contactfetchrequest.h"
#include "logger.h"

////////////////////////////////////////////////////////////////////////////////////////////////////

#define REPORT_UNSUPPORTED_FUNCTION(error) do {\
    qctWarn(QString::fromLatin1("Method not implemented yet: %1"). \
            arg(QLatin1String(Q_FUNC_INFO))); \
    if (error) \
       *error = QContactManager::NotSupportedError; \
} while(0)

////////////////////////////////////////////////////////////////////////////////////////////////////

// added here as list was not working because of
// [warn ] QObject::connect: Cannot queue arguments of type 'QContactAbstractRequest::State'
// (Make sure 'QContactAbstractRequest::State' is registered using qRegisterMetaType().)
Q_DECLARE_METATYPE(QContactAbstractRequest::State)

////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////

QContactTelepathyEngineData::QContactTelepathyEngineData(const QMap<QString, QString>& parameters,
                                                     const QString& engineName, int engineVersion)
    : QSharedData()
    , m_engineName(engineName)
    , m_engineVersion(engineVersion)
    , m_parameters(parameters)
    , m_managerParameters(0)
{
    static const QMap<QString, QString> environmentParameters(readEnvironment());
    m_parameters.unite(environmentParameters);
    parseParameters();
}

QContactTelepathyEngineData::QContactTelepathyEngineData(const QContactTelepathyEngineData& other)
    : QSharedData(other)
    , m_engineName(other.m_engineName)
    , m_engineVersion(other.m_engineVersion)
    , m_requests() // don't copy the request workers to prevent double frees
{
}

QContactTelepathyEngineData::~QContactTelepathyEngineData()
{
    QMutexLocker locker(&m_mutex);
    RequestList workers(m_requests.values());
    m_requests.clear();
    locker.unlock(); // silence gcc

    qDeleteAll(workers);
}

QMap<QString, QString>
QContactTelepathyEngineData::readEnvironment()
{
    const QStringList tokenList(QProcessEnvironment::systemEnvironment().
                                value(QLatin1String("QT_CONTACTS_TELEPATHY")).
                                replace(QLatin1Char(' '), QString()).
                                split(QLatin1String(";")));

    QMap<QString, QString> parameters;

    foreach(const QString &token, tokenList) {
        const int eq = token.indexOf(QLatin1Char('='));
        const QString key((eq >= 0 ? token.left(eq) : token).trimmed());
        const QString value((eq >= 0 ? token.mid(eq + 1).trimmed() : QString()));

        if (not key.isEmpty()) {
            parameters.insert(key, value);
        }
    }

    return parameters;
}

void
QContactTelepathyEngineData::parseParameters()
{
    QContactTelepathyEngine::DebugFlags debugFlags;
    QContactTelepathyEngine::DebugFlags queryBuilderFlags;

    for(QMap<QString, QString>::ConstIterator i(m_parameters.begin()); i != m_parameters.end(); ++i) {
        if (QLatin1String("debug") == i.key()) {
            const QSet<QString> values = i.value().split(QLatin1Char(',')).toSet();
            const bool all(values.contains(QLatin1String("all")));

            if (all || values.contains(QLatin1String("notes"))) {
                debugFlags |= QContactTelepathyEngine::ShowNotes;
            }
            if (all || values.contains(QLatin1String("timing"))) {
                debugFlags |= QContactTelepathyEngine::ShowTiming;
            }
            if (all || values.contains(QLatin1String("signals"))) {
                debugFlags |= QContactTelepathyEngine::ShowSignals;
            }
        } else if (QLatin1String("com.nokia.qt.mobility.contacts.implementation.version") == i.key()) {
        } else if (QLatin1String("AccountPath") == i.key()) {
            m_accountPath = i.value();
        } else {
            qctWarn(QString::fromLatin1("Unknown parameter: %1").arg(i.key()));
        }
    }

    m_debugFlags = debugFlags | queryBuilderFlags;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Instances which manage contact engine instances
////////////////////////////////////////////////////////////////////////////////////////////////////

QContactTelepathyEngine::QContactTelepathyEngine(const QMap<QString, QString>& parameters,
                                             const QString& managerName, int interfaceVersion)
{
    d = new QContactTelepathyEngineData(parameters, managerName, interfaceVersion);
    init();
}

QContactTelepathyEngine::QContactTelepathyEngine(const QContactTelepathyEngine& other)
    : d(other.d)
{
    d.detach();
    init();
}

void
QContactTelepathyEngine::init()
{
    Tp::registerTypes();
}


QContactTelepathyEngine::~QContactTelepathyEngine()
{
    // No need to lock m_requests since it's thread local.
    QContactTelepathyEngineData::RequestMap::Iterator it = d->m_requests.begin();
    while (it != d->m_requests.end()) {
        QContactAbstractRequest *request = static_cast<QContactAbstractRequest*>(it.key());
        QTelepathyAbstractRequest *trequest = it.value();

        if (trequest->engine() != this) {
            ++it;
            continue;
        }

        updateRequestState(request, QContactAbstractRequest::CanceledState);
        delete trequest;
        it = d->m_requests.erase(it);
    }
}

QContactTelepathyEngine&
QContactTelepathyEngine::operator=(const QContactTelepathyEngine& other)
{
    d = other.d;
    return *this;
}

QString
QContactTelepathyEngine::managerName() const
{
    return d->m_engineName;
}

int
QContactTelepathyEngine::managerVersion() const
{
    return d->m_engineVersion;
}

bool
QContactTelepathyEngine::hasFeature(QContactManager::ManagerFeature feature,
                                  const QString& contactType) const
{
    if (not supportedContactTypes().contains(contactType)) {
        return false;
    }

    switch (feature) {
    // FIXME    case QContactManager::SelfContact:
    //    return true;

    default:
        return false;
    }
}

bool
QContactTelepathyEngine::isFilterSupported(const QContactFilter& filter) const
{
    // FIXME - move the check to query builder. First implementation only remote search
    return QContactFilter::ContactDetailFilter == filter.type()
            && QContactDisplayLabel::DefinitionName == QContactDetailFilter(filter).detailDefinitionName();
}

QMap<QString, QContactDetailDefinition>
QContactTelepathyEngine::detailDefinitions(const QString& contactType, QContactManager::Error* error) const
{
    if (error) {
        *error = QContactManager::NoError;
    }
    QMap<QString, QMap<QString, QContactDetailDefinition> > defns = QContactManagerEngine::schemaDefinitions();

    // Remove the details we don't support
    defns[contactType].remove(QContactSyncTarget::DefinitionName);
    defns[contactType].remove(QContactGeoLocation::DefinitionName);
    defns[contactType].remove(QContactTimestamp::DefinitionName);
    defns[contactType].remove(QContactGuid::DefinitionName);
    defns[contactType].remove(QContactGender::DefinitionName);

    // Simple anniversarys
    defns[contactType][QContactAnniversary::DefinitionName].removeField(QContactAnniversary::FieldCalendarId);
    defns[contactType][QContactAnniversary::DefinitionName].removeField(QContactAnniversary::FieldEvent);
    defns[contactType][QContactAnniversary::DefinitionName].removeField(QContactAnniversary::FieldSubType);

    // No logo for organisation
    defns[contactType][QContactOrganization::DefinitionName].removeField(QContactOrganization::FieldLogoUrl);

    // No subtypes for these details
    defns[contactType][QContactUrl::DefinitionName].removeField(QContactUrl::FieldSubType);

    // No contexts for these details
    defns[contactType][QContactAvatar::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactAnniversary::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactBirthday::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactName::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactNickname::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactOrganization::DefinitionName].removeField(QContactDetail::FieldContext);
    defns[contactType][QContactUrl::DefinitionName].removeField(QContactDetail::FieldContext);

    // Simple phone number types (non multiple)
    // defns[QContactPhoneNumber::DefinitionName].fields()[QContactPhoneNumber::FieldSubTypes].dataType = QVariant::String; // XXX doesn't work

    QMap<QString, QContactDetailFieldDefinition> tmp = defns[contactType][QContactPhoneNumber::DefinitionName].fields();

    tmp[QContactPhoneNumber::FieldSubTypes].allowableValues().removeAll(QString(QLatin1String(QContactPhoneNumber::SubTypeBulletinBoardSystem)));
    tmp[QContactPhoneNumber::FieldSubTypes].allowableValues().removeAll(QString(QLatin1String(QContactPhoneNumber::SubTypeLandline)));
    tmp[QContactPhoneNumber::FieldSubTypes].allowableValues().removeAll(QString(QLatin1String(QContactPhoneNumber::SubTypeMessagingCapable)));
    tmp[QContactPhoneNumber::FieldSubTypes].allowableValues().removeAll(QString(QLatin1String(QContactPhoneNumber::SubTypeModem)));
    tmp[QContactPhoneNumber::FieldSubTypes].allowableValues().removeAll(QString(QLatin1String(QContactPhoneNumber::SubTypeVideo)));
    defns[contactType][QContactPhoneNumber::DefinitionName].setFields(tmp);

    return defns[contactType];
}

QList<QVariant::Type>
QContactTelepathyEngine::supportedDataTypes() const
{
    QList<QVariant::Type> st;
    st.append(QVariant::String);
    st.append(QVariant::Int);
    st.append(QVariant::UInt);
    st.append(QVariant::Double);
    st.append(QVariant::Date);
    st.append(QVariant::DateTime);

    return st;
}

QStringList
QContactTelepathyEngine::supportedContactTypes() const
{
    return QStringList() << QContactType::TypeContact;
}

const QContactTelepathyEngine::DebugFlags &
QContactTelepathyEngine::debugFlags() const
{
    return d->m_debugFlags;
}

bool
QContactTelepathyEngine::hasDebugFlag(DebugFlag flag) const
{
    return d->m_debugFlags.testFlag(flag);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Asynchronous data access methods
////////////////////////////////////////////////////////////////////////////////////////////////////


bool
QContactTelepathyEngine::startRequest(QContactAbstractRequest *request)
{
    // ensure old worker got destroyed when request gets reused
    requestDestroyed(request);

    QTelepathyAbstractRequest *worker = 0;
    QElapsedTimer t; t.start();

    // choose proper request implementation
    switch(request->type())
    {
    case QContactAbstractRequest::ContactFetchRequest:
        worker = new QTelepathyContactFetchRequest(request, this);
        break;

    default:
        break;
    }

    if (0 == worker) {
        qctWarn(QString::fromLatin1("Unsupported request type: %1").
                arg(QLatin1String(request->metaObject()->className())));
        return false;
    }

    if (hasDebugFlag(ShowNotes)) {
        qDebug() << Q_FUNC_INFO << "time elapsed while constructing request workers:"<< request << t.elapsed();
    }

    // the contacts engine must bring requests into Activate state as soon as it is dealing with them
    // the (secret) state machine only permits this transitions Inactive > Active > Canceled/Finished
    // because in case of an error the state can go to Canceled/Finished, set to Active now
    // before starting the worker
    updateRequestState(request, QContactAbstractRequest::ActiveState);

    if (hasDebugFlag(ShowNotes)) {
        qDebug() << Q_FUNC_INFO << "running" << worker->metaObject()->className();
    }

    // XXX The unit tests directly access the engine. Therefore requests created by our unit
    // tests don't have a manager attached. This prevents ~QContactAbstractRequest() to call our
    // engine's requestDestroyed(QContactAbstractRequest*) method, which results in memory leaks
    // within our our unit tests. To prevent those leaks we watch the request's destroyed()
    // signal and forward those signals to requestDestroyed(QContactAbstractRequest*) when
    // a request without contact manager is found.
    //
    // XXX This all of course is an ugly workaround. We probably should change the unit tests
    // to create use QContactManager accessing the engine via a static plugin.
    if (0 == request->manager()) {
        connect(request, SIGNAL(destroyed(QObject*)),
                SLOT(onRequestDestroyed(QObject*)));
    }

    // store the request and start it
    QMutexLocker locker(&d->m_mutex);
    d->m_requests.insert(request, worker);
    locker.unlock(); // silence gcc

    bool success = worker->start();

    if (hasDebugFlag(ShowNotes)) {
        qDebug() << Q_FUNC_INFO << "time elapsed until start returned:" << request << t.elapsed();
    }

    return success;
}

/*! \reimp */
void
QContactTelepathyEngine::requestDestroyed(QContactAbstractRequest *req)
{
    if (0 != req) {
        // the worker itself must be deleted outside of the mutex to prevent
        // deadlocks if the worker should use additional internal requests
        // for doing its job
        QMutexLocker locker(&d->m_mutex);
        QTelepathyAbstractRequest *worker(d->m_requests.take(req));
        locker.unlock(); // silence gcc

        if (hasDebugFlag(ShowNotes)) {
            qDebug()
                    << metaObject()->className() << ": request destroyed:"
                    << (void* ) req << worker;
        }

        delete worker;
    }
}

void
QContactTelepathyEngine::onRequestDestroyed(QObject *req)
{
    // dynamic_cast<> does not work at this point (has a 0 result, or even crashes) because the
    // derived parts of the class have already been destroyed by the time the base QObject's
    // destructor has emitted this signal. So we do a regular static case, because
    // requestDestroyed(QContactAbstractRequest*) just compares the pointer value anyway.
    requestDestroyed(static_cast<QContactAbstractRequest *>(req));
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// Unsupported functions
////////////////////////////////////////////////////////////////////////////////////////////////////

QContact
QContactTelepathyEngine::compatibleContact(const QContact&, QContactManager::Error* error) const
{
    REPORT_UNSUPPORTED_FUNCTION(error);
    return QContact();
}

bool
QContactTelepathyEngine::validateContact(const QContact&, QContactManager::Error* error) const
{
    REPORT_UNSUPPORTED_FUNCTION(error);
    return false;
}

bool
QContactTelepathyEngine::validateDefinition(const QContactDetailDefinition&,
                                          QContactManager::Error* error) const
{
    REPORT_UNSUPPORTED_FUNCTION(error);
    return false;
}

bool
QContactTelepathyEngine::saveDetailDefinition(const QContactDetailDefinition&, const QString&,
                                            QContactManager::Error* error)
{
    REPORT_UNSUPPORTED_FUNCTION(error);
    return false;
}

bool
QContactTelepathyEngine::removeDetailDefinition(const QString&, const QString&,
                                              QContactManager::Error* error)
{
    REPORT_UNSUPPORTED_FUNCTION(error);
    return false;
}

bool
QContactTelepathyEngine::cancelRequest(QContactAbstractRequest *request)
{
    if (0 == request) {
        return false;
    }

    updateRequestState(request, QContactAbstractRequest::CanceledState);

    QTelepathyAbstractRequest *worker = d->m_requests.take(request);

    if (0 == worker) {
        return false;
    }

    delete worker;
    return true;
}

QString
QContactTelepathyEngine::accountPath() const
{
    return d->m_accountPath;
}
