/**************************************************************************
**
** This file is part of libqtcontacts-telepathy
**
** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
**
** Contact: Nokia Corporation (info@qt.nokia.com)
**
**
** GNU Lesser General Public License Usage
**
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this file.
** Please review the following information to ensure the GNU Lesser General
** Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** Other Usage
**
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at info@qt.nokia.com.
**
**************************************************************************/


#include <qtcontacts.h>
#include <qcontactdetails.h>

#include "telepathy2qcontactconvertor.h"

///////////////////////////////////////////////////////////////////////////////

QMultiHash<QContactLocalId, TelepathyContactId > Telepathy2QContactConvertor::m_contactId2TelepathyId;
QMultiHash<QString, QContactLocalId> Telepathy2QContactConvertor::m_telepathyId2ContactId;

///////////////////////////////////////////////////////////////////////////////

// multiple ids might map to one QContactLocalId (with qHash() possibility exists so not ignoring it, at least not here)
QList< TelepathyContactId > Telepathy2QContactConvertor::contactId2TelepathyIds(QContactLocalId contactId)
{
    return m_contactId2TelepathyId.values(contactId);
}

QContactLocalId Telepathy2QContactConvertor::telepathyId2ContactId(const TelepathyContactId &id)
{
    QContactLocalId localId = qHash(id);
    if (not m_contactId2TelepathyId.contains(localId, id)) {
        m_contactId2TelepathyId.insert(localId, id);
    }
    return localId;
}

QString Telepathy2QContactConvertor::contactFieldContextConversion(const Tp::ContactInfoField &field)
{
    QString context(QLatin1String("Other"));
    static QHash<QString, QString> knownTypes;
    if (knownTypes.isEmpty()) {
        knownTypes.insert (QLatin1String("work"), QLatin1String("Work"));
        knownTypes.insert (QLatin1String("home"), QLatin1String("Home"));
    }

    Q_FOREACH (const QString &parameter, field.parameters) {
        if (!parameter.startsWith(QLatin1String("type="))) {
            continue;
        }

        const QString str = parameter.mid(5);
        if (knownTypes.contains(str)) {
            context = knownTypes[str];
            break;
        }
    }
    return context;
}

void
Telepathy2QContactConvertor::contactFieldConversionAppendToContact(const Tp::ContactInfoField &field, QContact &contact)
{
    QContactDetail detail;
    if (!field.fieldName.compare(QLatin1String("tel"))) {
        detail = QContactPhoneNumber();
        detail.setContexts(contactFieldContextConversion(field));
        detail.setValue(QContactPhoneNumber::FieldNumber, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("adr"))) {
        detail = QContactAddress();
        detail.setContexts(contactFieldContextConversion(field));
        detail.setValue(QContactAddress::FieldPostOfficeBox, field.fieldValue.at(0));
        /* at(1) extendedAddress */
        detail.setValue(QContactAddress::FieldStreet, field.fieldValue.at(2));
        detail.setValue(QContactAddress::FieldLocality, field.fieldValue.at(3));
        detail.setValue(QContactAddress::FieldRegion, field.fieldValue.at(4));
        detail.setValue(QContactAddress::FieldPostcode, field.fieldValue.at(5));
        detail.setValue(QContactAddress::FieldCountry, field.fieldValue.at(6));
    }
    else if (!field.fieldName.compare(QLatin1String("email"))) {
        detail = QContactEmailAddress();
        detail.setContexts(contactFieldContextConversion(field));
        detail.setValue(QContactEmailAddress::FieldEmailAddress, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("url"))) {
        detail = QContactUrl();
        detail.setContexts(contactFieldContextConversion(field));
        detail.setValue(QContactUrl::FieldUrl, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("title"))) {
        // take (if) one already existing
        detail = contact.detail<QContactOrganization>();
        detail.setValue(QContactOrganization::FieldTitle, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("role"))) {
        // take (if) one already existing
        detail = contact.detail<QContactOrganization>();
        detail.setValue(QContactOrganization::FieldRole, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("note")) || !field.fieldName.compare(QLatin1String("desc"))) {
        detail = QContactNote();
        detail.setValue(QContactNote::FieldNote, field.fieldValue.at(0));
    }
    else if (!field.fieldName.compare(QLatin1String("bday"))) {
        /* VCard spec allows only ISO 8601, but most IM clients allows
         * any string. */
        QDate date = QDate::fromString(field.fieldValue[0], QLatin1String("yyyy-MM-dd"));
        if (!date.isValid()) {
            date = QDate::fromString(field.fieldValue[0], QLatin1String("yyyyMMdd"));
        }

        if (date.isValid()) {
            detail.setValue(QContactBirthday::FieldBirthday, date);
        } else {
            qctWarn(QString::fromLatin1("Unsupported bday format:%1").arg(field.fieldValue[0]));
        }
    }
    if (!detail.isEmpty()) {
        contact.saveDetail(&detail);
    }
}

bool Telepathy2QContactConvertor::convertTelepathyContacts2QContacts(const TelepathyContacts &input,
        QList<QContact> &output, const QString &accountPath, const QContactManagerEngine * const engine)
{
    output.clear();
    // prepare working data
    QContactId contactId;
    if(engine) {
        contactId.setManagerUri(engine->managerUri());
    }

    QHashIterator<Tp::ContactPtr, Tp::Contact::InfoFields> i(input);
    while(i.hasNext()) {
        i.next();
        Tp::ContactPtr cptr = i.key();

        QContact contact;

        contactId.setLocalId(telepathyId2ContactId(TelepathyContactId(accountPath, cptr->id())));
        contact.setId(contactId);

        QContactOnlineAccount account;
        account.setAccountUri(cptr->id());
        account.setValue(QLatin1String("AccountPath"), accountPath);

        QContactPresence presence;
        // presence.setNickname(cptr->alias());
        Tp::Presence tpPresence(cptr->presence());
        presence.setPresenceState(presenceConversion(tpPresence.type()));
        presence.setCustomMessage(tpPresence.statusMessage());
        presence.setDetailUri(QString(QLatin1String("presence:%1!%2")).
                arg(accountPath).
                arg(cptr->id()));
        account.setLinkedDetailUris(QStringList() << presence.detailUri());

        Tp::ContactInfoFieldList infoFieldsList= i.value().allFields();
        Q_FOREACH (const Tp::ContactInfoField &field, infoFieldsList) {
            if (field.fieldValue.count() == 0) {
                continue;
            }
            if (!field.fieldName.compare(QLatin1String("fn"))) {
                presence.setNickname(field.fieldValue.first());
            }
            contactFieldConversionAppendToContact(field, contact);
        }

        contact.saveDetail(&account);
        contact.saveDetail(&presence);

        output << contact;
    }

    if(output.isEmpty())
        return false;
    return true;
}

QContactPresence::PresenceState
Telepathy2QContactConvertor::presenceConversion(const Tp::ConnectionPresenceType &presenceType)
{
    switch (presenceType) {
        case Tp::ConnectionPresenceTypeUnset:
            return QContactPresence::PresenceUnknown;
        case Tp::ConnectionPresenceTypeOffline:
            return QContactPresence::PresenceOffline;
        case Tp::ConnectionPresenceTypeAvailable:
            return QContactPresence::PresenceAvailable;
        case Tp::ConnectionPresenceTypeAway:
            return QContactPresence::PresenceAway;
        case Tp::ConnectionPresenceTypeExtendedAway:
            return QContactPresence::PresenceExtendedAway;
        case Tp::ConnectionPresenceTypeHidden:
            return QContactPresence::PresenceHidden;
        case Tp::ConnectionPresenceTypeBusy:
            return QContactPresence::PresenceBusy;
        default:
            return QContactPresence::PresenceUnknown;
    }
}
