/**************************************************************************
**
** This file is part of libqtcontacts-telepathy
**
** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
**
** Contact: Nokia Corporation (info@qt.nokia.com)
**
**
** GNU Lesser General Public License Usage
**
** This file may be used under the terms of the GNU Lesser General Public
** License version 2.1 as published by the Free Software Foundation and
** appearing in the file LICENSE.LGPL included in the packaging of this file.
** Please review the following information to ensure the GNU Lesser General
** Public License version 2.1 requirements will be met:
** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain additional
** rights. These rights are described in the Nokia Qt LGPL Exception
** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
**
** Other Usage
**
** Alternatively, this file may be used in accordance with the terms and
** conditions contained in a signed written agreement between you and Nokia.
**
** If you have questions regarding the use of this file, please contact
** Nokia at info@qt.nokia.com.
**
**************************************************************************/

#include <QCoreApplication>
#include <QTextStream>

#include <TelepathyQt4/AccountSet>
#include <TelepathyQt4/AccountPropertyFilter>
#include <TelepathyQt4/AndFilter>
#include <TelepathyQt4/NotFilter>

#include "qct-toolclient.h"

///////////////////////////////////////////////////////////////////////////////////////////////////

void
reportError(const QString &error)
{
    QTextStream(stdout) << error << endl;
    QCoreApplication::exit(EXIT_FAILURE);
}

void
QCTToolClient::listAccounts()
{
    const QDBusConnection &bus = QDBusConnection::sessionBus();
    Tp::AccountFactoryPtr accountFactory = Tp::AccountFactory::create(bus,
            Tp::Features() << Tp::Account::FeatureCore
                           << Tp::Account::FeatureAvatar
                           << Tp::Account::FeatureCapabilities);
    Tp::ConnectionFactoryPtr connectionFactory = Tp::ConnectionFactory::create(bus,
            Tp::Features() << Tp::Connection::FeatureCore
                           << Tp::Connection::FeatureRoster);
    Tp::ChannelFactoryPtr channelFactory = Tp::ChannelFactory::create(bus);
    Tp::ContactFactoryPtr contactFactory = Tp::ContactFactory::create(
            Tp::Features() << Tp::Contact::FeatureAlias
                           << Tp::Contact::FeatureAvatarToken
                           << Tp::Contact::FeatureAvatarData
                           << Tp::Contact::FeatureSimplePresence
                           << Tp::Contact::FeatureInfo
                           << Tp::Contact::FeatureLocation
                           << Tp::Contact::FeatureCapabilities);
    m_accountManager = Tp::AccountManager::create(bus, accountFactory, connectionFactory,
            channelFactory, contactFactory);

    // Wait for account manager to become ready
    connect(m_accountManager->becomeReady(Tp::AccountManager::FeatureCore),
            SIGNAL(finished(Tp::PendingOperation*)),
            SLOT(onAccountManagerReady(Tp::PendingOperation*)));
}

void
QCTToolClient::search(const QString &accountPath, const QString &searchPattern)
{
    m_accountPath = accountPath;
    if (accountPath.isEmpty()) {
        reportError(QLatin1String("AccountPath was not supplied as parameter to QContactManager(\"telepathy\", parameters).\n"
                    "e.g.:\n"
                    "QMap<QString, QString>& parameters;\n"
                    "parameters[QLatin1String(\"AccountPath\")] = accountPath;\n"
                    "QContactManager *mTelepathyManager = new QContactManager(QLatin1String(\"telepathy\"), parameters);\n"
                    "Not able to search for contacts if account path is not supplied.\n"));

    }
    Tp::Features wantedFeatures;

    wantedFeatures << Tp::Account::FeatureCore
                   << Tp::Account::FeatureProfile
                   << Tp::Account::FeatureProtocolInfo;

    m_account = Tp::Account::create(QLatin1String(TELEPATHY_ACCOUNT_MANAGER_BUS_NAME), accountPath);
    m_searchPattern = searchPattern;
    getChannel();
}

void
QCTToolClient::getChannel()
{
    Tp::PendingChannel *pendingChannel = m_account->createAndHandleContactSearch(QString());
    connect(pendingChannel, SIGNAL(finished(Tp::PendingOperation *)),
            this, SLOT(gotChannel(Tp::PendingOperation *)));
}

void
QCTToolClient::gotChannel(Tp::PendingOperation*op)
{
    if(op->isError()) {
        reportError(QString::fromLatin1("Search connection channel not available:%1-%2 for account:%3. Not able to search.").
                arg(op->errorName()).
                arg(op->errorMessage()).
                arg(m_accountPath));
    } else {
        Tp::PendingChannel *pendingChannel = qobject_cast<Tp::PendingChannel *>(op);
        m_channel = pendingChannel->channel();
        m_contactSearchChannel = Tp::ContactSearchChannelPtr::qObjectCast(m_channel);

        connect(m_contactSearchChannel->becomeReady(Tp::ContactSearchChannel::FeatureCore),SIGNAL(finished (Tp::PendingOperation*)),
                     SLOT(onChannelReady(Tp::PendingOperation*)));
    }
 }

void
QCTToolClient::onChannelReady(Tp::PendingOperation *op)
{
    if (op->isError()) {
        reportError(QString::fromLatin1("Search connection channel not ready:%1-%2 for account:%3. Not able to search.").
                arg(op->errorName()).
                arg(op->errorMessage()).
                arg(m_accountPath));
    } else {
        m_searchOperation = m_contactSearchChannel->search(QLatin1String(""), m_searchPattern);

        connect(m_contactSearchChannel.data(), SIGNAL(searchStateChanged(Tp::ChannelContactSearchState ,const QString &, const Tp::ContactSearchChannel::SearchStateChangeDetails &)),
                                              SLOT(onSearchStateChanged(Tp::ChannelContactSearchState ,const QString &, const Tp::ContactSearchChannel::SearchStateChangeDetails &)));

        connect(m_contactSearchChannel.data(), SIGNAL(searchResultReceived(const Tp::ContactSearchChannel::SearchResult &)),
                                              SLOT(onSearchResultReceived(const Tp::ContactSearchChannel::SearchResult&)));
    }
}

void
QCTToolClient::onSearchStateChanged(Tp::ChannelContactSearchState state,const QString &error, const Tp::ContactSearchChannel::SearchStateChangeDetails &details)
{
    Q_UNUSED(details)
    if(state==Tp::ChannelContactSearchStateNotStarted) {
        reportError("Search not started");
    } else if (state==Tp::ChannelContactSearchStateInProgress) {
        QTextStream(stdout) << Q_FUNC_INFO << __LINE__ << "Tp::ChannelContactSearchStateInProgress" << endl;
    } else if (state==Tp::ChannelContactSearchStateCompleted) {
        QTextStream(stdout) << Q_FUNC_INFO << __LINE__ << "Tp::ChannelContactSearchStateCompleted" << endl;
        m_contactSearchChannel->requestClose();
        QCoreApplication::exit(EXIT_SUCCESS);
    } else if (state==Tp::ChannelContactSearchStateFailed) {
        reportError(QString::fromLatin1("Search failed:%1 - for account:%3. Not able to search.").
                arg(error).
                arg(m_accountPath));
    }
}

void
QCTToolClient::onSearchResultReceived(const Tp::ContactSearchChannel::SearchResult &result)
{
    QTextStream(stdout) << Q_FUNC_INFO << __LINE__ << QString::fromLatin1("search returned %1 contacts:").arg(result.size()) << endl;

    QHashIterator<Tp::ContactPtr, Tp::Contact::InfoFields> i(result);
    while(i.hasNext()) {
        i.next();
        QTextStream(stdout) << i.key()->id() << endl;
    }
}

void
QCTToolClient::onAccountManagerReady(Tp::PendingOperation *op)
{
    if (op->isError()) {
        reportError(QString::fromLatin1("Could not make account manager ready:%1 - %2")
                    .arg(op->errorName())
                    .arg(op->errorMessage()));
    }

    Tp::AccountPropertyFilterPtr filter1 = Tp::AccountPropertyFilter::create();
    filter1->addProperty(QString::fromLatin1("valid"), true);
    filter1->addProperty(QString::fromLatin1("hasBeenOnline"), true);

    Tp::AccountPropertyFilterPtr normalizedNameFilter = Tp::AccountPropertyFilter::create();
    normalizedNameFilter->addProperty(QString::fromLatin1("normalizedName"), QString());
    Tp::AccountFilterPtr filter2 = Tp::NotFilter<Tp::Account>::create(normalizedNameFilter);

    Tp::AccountFilterPtr filter = Tp::AndFilter<Tp::Account>::create(
            QList<Tp::AccountFilterConstPtr>() << filter1 << filter2);

    Tp::AccountSetPtr accountSet = m_accountManager->filterAccounts(filter);

    Q_FOREACH (const Tp::AccountPtr &account, accountSet->accounts()) {
        QTextStream(stdout) << account->objectPath() << endl;
    }
    QCoreApplication::exit(EXIT_SUCCESS);
}
